open Printf
open AbstractSyntax

(* ------------------------------------------------------------------------- *)

(* Read the file and parse the program. *)

let Program (_, main) as program =
  let channel = open_in Settings.filename in
  let lexbuf = Lexing.from_channel channel in
  lexbuf.Lexing.lex_curr_p <-
      { 
	Lexing.pos_fname = Settings.filename; 
	Lexing.pos_lnum  = 1;
	Lexing.pos_bol   = 0; 
	Lexing.pos_cnum  = 0
      };
  try
    let program = Parser.program Lexer.main lexbuf in
    close_in channel;
    program
  with Parser.Error ->
    Error.error
      [ Some (Lexing.lexeme_start_p lexbuf, Lexing.lexeme_end_p lexbuf) ]
      "Syntax error."

(* ------------------------------------------------------------------------- *)

(* Check that the program is well-formed and extract the data constructor
   table. *)

let dcenv =
  Wf.check_program program

(* ------------------------------------------------------------------------- *)

(* Perform type inference and print the program's type. *)

let t : Unification.term =
  let module P = Generator.Run (struct
    let dcenv = dcenv
    let main = main
  end) in
  try
    let mgu = Unification.mgu P.problem in
    try
      Unification.Var.Map.find P.root mgu
    with Not_found ->
      assert false
  with
  | Unification.Mismatch ->
      Error.error [] "Unification error (mismatch)."
  | Unification.OccurCheck ->
      Error.error [] "Unification error (occur check)."

let () =
  printf "*** Type:\n";
  Print.wrap Unification.print stdout t;
  printf "\n%!"

(* ------------------------------------------------------------------------- *)

(* Execute the program *)

let _ =
  printf "*** Interpretation:\n";
  Interpreter.eval_program program

(* ------------------------------------------------------------------------- *)

(* Compile the program, print the generated code, and execute it *)

let _ =
  try
    let code = Compiler.compile_program program in
    printf "*** Compiled code:\n";
    Machine.print_program code;
    printf "*** Execution of the compiled code:\n";
    Machine.execute_program code
  with Compiler.CompilationError msg ->
    printf "*** Compilation error: %s\n" msg






